## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import unittest

import gevent
from gevent import coros
from gevent import event
from gevent import socket
import mock

from tendril import application
from tendril import connection
from tendril import framers
from tendril import manager
from tendril import tcp


class TestException(Exception):
    pass


class TestTCPTendril(unittest.TestCase):
    def setUp(self):
        self.sock = mock.Mock(**{
            'getsockname.return_value': ('127.0.0.1', 8080),
            'getpeername.return_value': ('127.0.0.2', 8880),
        })

    def test_init_withremote(self):
        tend = tcp.TCPTendril('manager', self.sock, ('127.0.0.2', 8880))

        self.assertEqual(tend.local_addr, ('127.0.0.1', 8080))
        self.assertEqual(tend.remote_addr, ('127.0.0.2', 8880))
        self.assertIsInstance(tend._recv_framer, framers.LineFramer)
        self.assertIsInstance(tend._send_framer, framers.LineFramer)
        self.assertEqual(id(tend._sock), id(self.sock))
        self.assertIsInstance(tend._sendbuf_event, event.Event)
        self.assertEqual(tend._sendbuf, '')
        self.assertEqual(tend._recv_thread, None)
        self.assertEqual(tend._send_thread, None)
        self.assertEqual(tend._recv_lock, None)
        self.assertEqual(tend._send_lock, None)
        self.sock.getsockname.assert_called_once_with()
        self.assertFalse(self.sock.getpeername.called)

    def test_init_noremote(self):
        tend = tcp.TCPTendril('manager', self.sock)

        self.assertEqual(tend.local_addr, ('127.0.0.1', 8080))
        self.assertEqual(tend.remote_addr, ('127.0.0.2', 8880))
        self.assertIsInstance(tend._recv_framer, framers.LineFramer)
        self.assertIsInstance(tend._send_framer, framers.LineFramer)
        self.assertEqual(id(tend._sock), id(self.sock))
        self.assertIsInstance(tend._sendbuf_event, event.Event)
        self.assertEqual(tend._sendbuf, '')
        self.assertEqual(tend._recv_thread, None)
        self.assertEqual(tend._send_thread, None)
        self.assertEqual(tend._recv_lock, None)
        self.assertEqual(tend._send_lock, None)
        self.sock.getsockname.assert_called_once_with()
        self.sock.getpeername.assert_called_once_with()

    @mock.patch.object(gevent, 'spawn')
    def test_start(self, mock_spawn):
        recv_thread = mock.Mock()
        send_thread = mock.Mock()
        mock_spawn.side_effect = [recv_thread, send_thread]

        tend = tcp.TCPTendril('manager', self.sock)
        tend._start()

        self.assertIsInstance(tend._recv_lock, coros.Semaphore)
        self.assertIsInstance(tend._send_lock, coros.Semaphore)
        mock_spawn.assert_has_calls([mock.call(tend._recv),
                                     mock.call(tend._send)])
        self.assertEqual(id(tend._recv_thread), id(recv_thread))
        self.assertEqual(id(tend._send_thread), id(send_thread))
        recv_thread.link.assert_called_once_with(tend._thread_error)
        send_thread.link.assert_called_once_with(tend._thread_error)

    @mock.patch.object(connection.Tendril, 'close')
    @mock.patch.object(tcp.TCPTendril, '_recv_frameify')
    @mock.patch.object(tcp.TCPTendril, 'closed')
    @mock.patch.object(gevent, 'sleep')
    def test_recv(self, mock_sleep, mock_closed, mock_recv_frameify,
                  mock_close):
        self.sock.recv.side_effect = ['frame 1', 'frame 2', '']
        tend = tcp.TCPTendril('manager', self.sock)
        tend._recv_lock = mock.Mock()
        send_thread = mock.Mock()
        tend._send_thread = send_thread

        with self.assertRaises(gevent.GreenletExit):
            tend._recv()

        self.assertEqual(tend._recv_lock.method_calls, [
            mock.call.release(), mock.call.acquire(),
            mock.call.release(), mock.call.acquire(),
            mock.call.release(), mock.call.acquire(),
        ])
        mock_sleep.assert_has_calls([mock.call(), mock.call(), mock.call()])
        self.sock.recv.assert_has_calls([mock.call(4096), mock.call(4096),
                                         mock.call(4096)])
        send_thread.kill.assert_called_once_with()
        self.assertEqual(tend._send_thread, None)
        self.sock.close.assert_called_once_with()
        self.assertEqual(tend._sock, None)
        mock_close.assert_called_once_with()
        mock_closed.assert_called_once_with()
        mock_recv_frameify.assert_has_calls([mock.call('frame 1'),
                                             mock.call('frame 2')])

    @mock.patch.object(connection.Tendril, 'close')
    @mock.patch.object(tcp.TCPTendril, '_recv_frameify')
    @mock.patch.object(tcp.TCPTendril, 'closed')
    @mock.patch.object(gevent, 'sleep')
    def test_recv_altbufsize(self, mock_sleep, mock_closed, mock_recv_frameify,
                             mock_close):
        self.sock.recv.side_effect = ['frame 1', 'frame 2', '']
        tend = tcp.TCPTendril('manager', self.sock)
        tend.recv_bufsize = 1024
        tend._recv_lock = mock.Mock()
        send_thread = mock.Mock()
        tend._send_thread = send_thread

        with self.assertRaises(gevent.GreenletExit):
            tend._recv()

        self.assertEqual(tend._recv_lock.method_calls, [
            mock.call.release(), mock.call.acquire(),
            mock.call.release(), mock.call.acquire(),
            mock.call.release(), mock.call.acquire(),
        ])
        mock_sleep.assert_has_calls([mock.call(), mock.call(), mock.call()])
        self.sock.recv.assert_has_calls([mock.call(1024), mock.call(1024),
                                         mock.call(1024)])
        send_thread.kill.assert_called_once_with()
        self.assertEqual(tend._send_thread, None)
        self.sock.close.assert_called_once_with()
        self.assertEqual(tend._sock, None)
        mock_close.assert_called_once_with()
        mock_closed.assert_called_once_with()
        mock_recv_frameify.assert_has_calls([mock.call('frame 1'),
                                             mock.call('frame 2')])

    def test_send(self):
        self.sock.send.side_effect = [7, 7]
        tend = tcp.TCPTendril('manager', self.sock)
        self.sock.reset_mock()
        tend._sendbuf = 'frame 1frame 2'
        tend._send_lock = mock.Mock()
        tend._sendbuf_event = mock.Mock(**{
            'clear.side_effect': gevent.GreenletExit,
        })

        with self.assertRaises(gevent.GreenletExit):
            tend._send()

        self.assertEqual(tend._send_lock.method_calls, [
            mock.call.release(), mock.call.acquire(),
        ])
        self.assertEqual(tend._sendbuf_event.method_calls, [
            mock.call.wait(), mock.call.clear(),
        ])
        self.assertEqual(tend._sendbuf, '')
        self.sock.send.assert_has_calls([mock.call('frame 1frame 2'),
                                         mock.call('frame 2')])

    @mock.patch.object(tcp.TCPTendril, 'close')
    @mock.patch.object(tcp.TCPTendril, 'closed')
    def test_thread_error_successful(self, mock_closed, mock_close):
        thread = mock.Mock(**{'successful.return_value': True})
        tend = tcp.TCPTendril('manager', self.sock)
        tend._send_thread = mock.Mock()
        tend._recv_thread = mock.Mock()

        tend._thread_error(thread)

        mock_close.assert_called_once_with()
        self.assertEqual(mock_closed.call_count, 1)

        args = mock_closed.call_args[0]
        self.assertEqual(len(args), 1)
        self.assertIsInstance(args[0], socket.error)
        self.assertEqual(args[0][0], 'thread exited prematurely')

        # Ensure we didn't overwrite the threads
        self.assertNotEqual(tend._send_thread, None)
        self.assertNotEqual(tend._recv_thread, None)

    @mock.patch.object(tcp.TCPTendril, 'close')
    @mock.patch.object(tcp.TCPTendril, 'closed')
    def test_thread_error_greenletexit(self, mock_closed, mock_close):
        thread = mock.Mock(**{
            'successful.return_value': False,
            'exception': gevent.GreenletExit(),
        })
        tend = tcp.TCPTendril('manager', self.sock)
        tend._send_thread = mock.Mock()
        tend._recv_thread = mock.Mock()

        tend._thread_error(thread)

        mock_close.assert_called_once_with()
        self.assertFalse(mock_closed.called)

        # Ensure we didn't overwrite the threads
        self.assertNotEqual(tend._send_thread, None)
        self.assertNotEqual(tend._recv_thread, None)

    @mock.patch.object(tcp.TCPTendril, 'close')
    @mock.patch.object(tcp.TCPTendril, 'closed')
    def test_thread_error_exception(self, mock_closed, mock_close):
        thread = mock.Mock(**{
            'successful.return_value': False,
            'exception': TestException(),
        })
        tend = tcp.TCPTendril('manager', self.sock)
        tend._send_thread = mock.Mock()
        tend._recv_thread = mock.Mock()

        tend._thread_error(thread)

        mock_close.assert_called_once_with()
        self.assertEqual(mock_closed.call_count, 1)

        args = mock_closed.call_args[0]
        self.assertEqual(len(args), 1)
        self.assertEqual(id(args[0]), id(thread.exception))

        # Ensure we didn't overwrite the threads
        self.assertNotEqual(tend._send_thread, None)
        self.assertNotEqual(tend._recv_thread, None)

    @mock.patch.object(tcp.TCPTendril, 'close')
    @mock.patch.object(tcp.TCPTendril, 'closed')
    def test_thread_error_sendthread(self, _mock_closed, _mock_close):
        thread = mock.Mock(**{'successful.return_value': True})
        tend = tcp.TCPTendril('manager', self.sock)
        tend._send_thread = thread
        tend._recv_thread = mock.Mock()

        tend._thread_error(thread)

        self.assertEqual(tend._send_thread, None)
        self.assertNotEqual(tend._recv_thread, None)

    @mock.patch.object(tcp.TCPTendril, 'close')
    @mock.patch.object(tcp.TCPTendril, 'closed')
    def test_thread_error_recvthread(self, _mock_closed, _mock_close):
        thread = mock.Mock(**{'successful.return_value': True})
        tend = tcp.TCPTendril('manager', self.sock)
        tend._send_thread = mock.Mock()
        tend._recv_thread = thread

        tend._thread_error(thread)

        self.assertNotEqual(tend._send_thread, None)
        self.assertEqual(tend._recv_thread, None)

    def test_wrap_nolock(self):
        wrapper = mock.Mock()
        tend = tcp.TCPTendril('manager', self.sock)
        tend._recv_lock = mock.Mock()
        tend._send_lock = mock.Mock()

        tend.wrap(wrapper)

        self.assertNotEqual(id(tend._sock), id(self.sock))
        wrapper.assert_called_once_with(self.sock)
        self.assertEqual(tend._recv_lock.mock_calls, [])
        self.assertEqual(tend._send_lock.mock_calls, [])

    def test_wrap_withlock(self):
        wrapper = mock.Mock()
        tend = tcp.TCPTendril('manager', self.sock)
        tend._recv_thread = mock.Mock()
        tend._send_thread = mock.Mock()
        tend._recv_lock = mock.Mock()
        tend._send_lock = mock.Mock()

        tend.wrap(wrapper)

        self.assertNotEqual(id(tend._sock), id(self.sock))
        wrapper.assert_called_once_with(self.sock)
        tend._recv_lock.assert_has_calls(
            [mock.call.acquire(), mock.call.release()])
        tend._send_lock.assert_has_calls(
            [mock.call.acquire(), mock.call.release()])

    def test_sock(self):
        tend = tcp.TCPTendril('manager', self.sock)

        self.assertEqual(id(tend.sock), id(self.sock))

    @mock.patch.object(connection.Tendril, '_send_streamify',
                       return_value='frame1:frame2')
    def test_send_frame(self, mock_send_streamify):
        tend = tcp.TCPTendril('manager', self.sock)
        tend._sendbuf_event = mock.Mock()

        tend.send_frame('a frame')

        mock_send_streamify.assert_called_once_with('a frame')
        tend._sendbuf_event.assert_has_calls([mock.call.set()])

    @mock.patch.object(connection.Tendril, 'close')
    def test_close_nothreads_nosock(self, mock_close):
        tend = tcp.TCPTendril('manager', self.sock)
        tend._sock = None

        tend.close()

        mock_close.assert_called_once_with()

    @mock.patch.object(connection.Tendril, 'close')
    def test_close_nothreads(self, mock_close):
        tend = tcp.TCPTendril('manager', self.sock)

        tend.close()

        mock_close.assert_called_once_with()
        self.assertEqual(tend._sock, None)
        self.sock.close.assert_called_once_with()

    @mock.patch.object(connection.Tendril, 'close')
    def test_close_threads(self, mock_close):
        tend = tcp.TCPTendril('manager', self.sock)
        recv_thread = mock.Mock()
        tend._recv_thread = recv_thread
        send_thread = mock.Mock()
        tend._send_thread = send_thread

        tend.close()

        mock_close.assert_called_once_with()
        self.assertEqual(tend._sock, None)
        self.assertEqual(tend._recv_thread, None)
        self.assertEqual(tend._send_thread, None)
        self.sock.close.assert_called_once_with()
        recv_thread.kill.assert_called_once_with()
        send_thread.kill.assert_called_once_with()


@mock.patch.dict(manager.TendrilManager._managers)
class TestTCPTendrilManager(unittest.TestCase):
    @mock.patch.object(socket, 'socket', return_value=mock.Mock())
    @mock.patch.object(manager.TendrilManager, 'connect')
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_connect_basic(self, mock_TCPTendril, mock_track_tendril,
                           mock_connect, mock_socket):
        acceptor = mock.Mock()
        manager = tcp.TCPTendrilManager()

        tend = manager.connect(('127.0.0.1', 8080), acceptor)

        mock_connect.assert_called_once_with(('127.0.0.1', 8080), acceptor,
                                             None)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.connect(('127.0.0.1', 8080)),
        ])
        mock_TCPTendril.assert_called_once_with(manager,
                                                mock_socket.return_value)
        acceptor.assert_called_once_with(mock_TCPTendril.return_value)
        mock_track_tendril.assert_called_once_with(
            mock_TCPTendril.return_value)
        mock_TCPTendril.return_value._start.assert_called_once_with()
        self.assertEqual(id(tend), id(mock_TCPTendril.return_value))

    @mock.patch.object(socket, 'socket', return_value=mock.Mock())
    @mock.patch.object(manager.TendrilManager, 'connect')
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_connect_wrapper(self, mock_TCPTendril, mock_track_tendril,
                             mock_connect, mock_socket):
        wrapped_sock = mock.Mock()
        wrapper = mock.Mock(return_value=wrapped_sock)
        acceptor = mock.Mock()
        manager = tcp.TCPTendrilManager(('127.0.0.2', 8880))

        tend = manager.connect(('127.0.0.1', 8080), acceptor, wrapper)

        mock_connect.assert_called_once_with(('127.0.0.1', 8080), acceptor,
                                             wrapper)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('127.0.0.2', 8880)),
            mock.call.connect(('127.0.0.1', 8080)),
        ])
        wrapper.assert_called_once_with(mock_socket.return_value)
        mock_TCPTendril.assert_called_once_with(manager, wrapped_sock)
        acceptor.assert_called_once_with(mock_TCPTendril.return_value)
        mock_track_tendril.assert_called_once_with(
            mock_TCPTendril.return_value)
        mock_TCPTendril.return_value._start.assert_called_once_with()
        self.assertEqual(id(tend), id(mock_TCPTendril.return_value))

    @mock.patch.object(socket, 'socket', return_value=mock.Mock())
    @mock.patch.object(manager.TendrilManager, 'connect')
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_connect_failure(self, mock_TCPTendril, mock_track_tendril,
                             mock_connect, mock_socket):
        acceptor = mock.Mock(side_effect=TestException())
        manager = tcp.TCPTendrilManager()

        with self.assertRaises(TestException):
            tend = manager.connect(('127.0.0.1', 8080), acceptor)

        mock_connect.assert_called_once_with(('127.0.0.1', 8080), acceptor,
                                             None)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.connect(('127.0.0.1', 8080)),
            mock.call.close(),
        ])
        mock_TCPTendril.assert_called_once_with(manager,
                                                mock_socket.return_value)
        acceptor.assert_called_once_with(mock_TCPTendril.return_value)
        self.assertFalse(mock_track_tendril.called)
        self.assertFalse(mock_TCPTendril.return_value._start.called)

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'close.side_effect': socket.error(),
    }))
    @mock.patch.object(manager.TendrilManager, 'connect')
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_connect_failure_closeerror(self, mock_TCPTendril,
                                        mock_track_tendril, mock_connect,
                                        mock_socket):
        acceptor = mock.Mock(side_effect=TestException())
        manager = tcp.TCPTendrilManager()

        with self.assertRaises(TestException):
            tend = manager.connect(('127.0.0.1', 8080), acceptor)

        mock_connect.assert_called_once_with(('127.0.0.1', 8080), acceptor,
                                             None)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.connect(('127.0.0.1', 8080)),
            mock.call.close(),
        ])
        mock_TCPTendril.assert_called_once_with(manager,
                                                mock_socket.return_value)
        acceptor.assert_called_once_with(mock_TCPTendril.return_value)
        self.assertFalse(mock_track_tendril.called)
        self.assertFalse(mock_TCPTendril.return_value._start.called)

    @mock.patch.object(socket, 'socket', return_value=mock.Mock())
    @mock.patch.object(manager.TendrilManager, 'connect')
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_connect_reject_connection(self, mock_TCPTendril,
                                       mock_track_tendril, mock_connect,
                                       mock_socket):
        acceptor = mock.Mock(side_effect=application.RejectConnection)
        manager = tcp.TCPTendrilManager()

        tend = manager.connect(('127.0.0.1', 8080), acceptor)

        self.assertEqual(tend, None)
        mock_connect.assert_called_once_with(('127.0.0.1', 8080), acceptor,
                                             None)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.connect(('127.0.0.1', 8080)),
            mock.call.close(),
        ])
        mock_TCPTendril.assert_called_once_with(manager,
                                                mock_socket.return_value)
        acceptor.assert_called_once_with(mock_TCPTendril.return_value)
        self.assertFalse(mock_track_tendril.called)
        self.assertFalse(mock_TCPTendril.return_value._start.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock())
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_listener_noacceptor(self, mock_TCPTendril, mock_track_tendril,
                                 mock_socket, mock_sleep):
        manager = tcp.TCPTendrilManager()

        with self.assertRaises(TestException):
            manager.listener(None, None)

        mock_sleep.assert_called_once_with(600)
        self.assertFalse(mock_socket.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'accept.side_effect': TestException(),
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_listener_creator(self, mock_TCPTendril, mock_track_tendril,
                              mock_socket, mock_sleep):
        acceptor = mock.Mock()
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(1024),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_TCPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'accept.side_effect': TestException(),
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    @mock.patch.object(tcp.TCPTendrilManager, 'backlog', 4096)
    def test_listener_creator_backlog(self, mock_TCPTendril,
                                      mock_track_tendril, mock_socket,
                                      mock_sleep):
        acceptor = mock.Mock()
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(4096),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_TCPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'accept.side_effect': gevent.GreenletExit(),
        'getsockname.return_value': ('127.0.0.1', 8080),
        'close.side_effect': socket.error(),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_listener_killed(self, mock_TCPTendril, mock_track_tendril,
                             mock_socket, mock_sleep):
        acceptor = mock.Mock()
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(gevent.GreenletExit):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(1024),
            mock.call.accept(),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_TCPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'accept.side_effect': TestException(),
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_listener_wrapper(self, mock_TCPTendril, mock_track_tendril,
                              mock_socket, mock_sleep):
        wrapped_sock = mock.Mock(**{
            'accept.side_effect': TestException(),
        })
        wrapper = mock.Mock(return_value=wrapped_sock)
        acceptor = mock.Mock()
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, wrapper)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
        ])
        wrapper.assert_called_once_with(mock_socket.return_value)
        wrapped_sock.assert_has_calls([
            mock.call.listen(1024),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_TCPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'accept.side_effect': TestException(),
        'getsockname.return_value': ('127.0.0.1', 8080),
        'listen.side_effect': TestException(),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_listener_nolisten(self, mock_TCPTendril, mock_track_tendril,
                               mock_socket, mock_sleep):
        acceptor = mock.Mock()
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(1024),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_TCPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril')
    def test_listener_acceptor(self, mock_TCPTendril, mock_track_tendril,
                               mock_socket, mock_sleep):
        clis = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_socket.return_value.accept.side_effect = [
            (clis[0], ('127.0.0.2', 8082)),
            (clis[1], ('127.0.0.3', 8083)),
            (clis[2], ('127.0.0.4', 8084)),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
        ]
        tendrils = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_TCPTendril.side_effect = tendrils[:]
        acceptor = mock.Mock(side_effect=[
            mock.Mock(),
            TestException(),
            mock.Mock(),
        ])
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(1024),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        mock_TCPTendril.assert_has_calls([
            mock.call(manager, clis[0], ('127.0.0.2', 8082)),
            mock.call(manager, clis[1], ('127.0.0.3', 8083)),
            mock.call(manager, clis[2], ('127.0.0.4', 8084)),
        ])
        acceptor.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[1]),
            mock.call(tendrils[2]),
        ])
        self.assertFalse(clis[0].close.called)
        clis[1].close.assert_called_once_with()
        self.assertFalse(clis[2].close.called)
        mock_track_tendril.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[2]),
        ])
        tendrils[0]._start.assert_called_once_with()
        self.assertFalse(tendrils[1]._start.called)
        tendrils[2]._start.assert_called_once_with()

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril')
    def test_listener_acceptor_reject_connection(self, mock_TCPTendril,
                                                 mock_track_tendril,
                                                 mock_socket, mock_sleep):
        clis = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_socket.return_value.accept.side_effect = [
            (clis[0], ('127.0.0.2', 8082)),
            (clis[1], ('127.0.0.3', 8083)),
            (clis[2], ('127.0.0.4', 8084)),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
        ]
        tendrils = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_TCPTendril.side_effect = tendrils[:]
        acceptor = mock.Mock(side_effect=application.RejectConnection())
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(1024),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        mock_TCPTendril.assert_has_calls([
            mock.call(manager, clis[0], ('127.0.0.2', 8082)),
            mock.call(manager, clis[1], ('127.0.0.3', 8083)),
            mock.call(manager, clis[2], ('127.0.0.4', 8084)),
        ])
        acceptor.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[1]),
            mock.call(tendrils[2]),
        ])
        clis[0].close.assert_called_once_with()
        clis[1].close.assert_called_once_with()
        clis[2].close.assert_called_once_with()
        self.assertFalse(mock_track_tendril.called)
        self.assertFalse(tendrils[0]._start.called)
        self.assertFalse(tendrils[1]._start.called)
        self.assertFalse(tendrils[2]._start.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'getsockname.return_value': ('127.0.0.1', 8080),
        'close.side_effect': socket.error(),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril')
    def test_listener_acceptor_badclose(self, mock_TCPTendril,
                                        mock_track_tendril, mock_socket,
                                        mock_sleep):
        clis = [
            mock.Mock(**{
                'close.side_effect': socket.error(),
            }),
            mock.Mock(**{
                'close.side_effect': socket.error(),
            }),
            mock.Mock(**{
                'close.side_effect': socket.error(),
            })
        ]
        mock_socket.return_value.accept.side_effect = [
            (clis[0], ('127.0.0.2', 8082)),
            (clis[1], ('127.0.0.3', 8083)),
            (clis[2], ('127.0.0.4', 8084)),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
        ]
        tendrils = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_TCPTendril.side_effect = tendrils[:]
        acceptor = mock.Mock(side_effect=[
            mock.Mock(),
            TestException(),
            mock.Mock(),
        ])
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(1024),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        mock_TCPTendril.assert_has_calls([
            mock.call(manager, clis[0], ('127.0.0.2', 8082)),
            mock.call(manager, clis[1], ('127.0.0.3', 8083)),
            mock.call(manager, clis[2], ('127.0.0.4', 8084)),
        ])
        acceptor.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[1]),
            mock.call(tendrils[2]),
        ])
        self.assertFalse(clis[0].close.called)
        clis[1].close.assert_called_once_with()
        self.assertFalse(clis[2].close.called)
        mock_track_tendril.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[2]),
        ])
        tendrils[0]._start.assert_called_once_with()
        self.assertFalse(tendrils[1]._start.called)
        tendrils[2]._start.assert_called_once_with()

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'accept.side_effect': TestException(),
        'getsockname.return_value': ('127.0.0.1', 8080),
        'listen.side_effect': TestException(),
        'close.side_effect': socket.error(),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril', return_value=mock.Mock())
    def test_listener_nolisten_noclose(self, mock_TCPTendril,
                                       mock_track_tendril, mock_socket,
                                       mock_sleep):
        acceptor = mock.Mock()
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(1024),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_TCPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(gevent, 'sleep', side_effect=TestException())
    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(tcp, 'TCPTendril')
    def test_listener_err_thresh(self, mock_TCPTendril, mock_track_tendril,
                                 mock_socket, mock_sleep):
        clis = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_socket.return_value.accept.side_effect = [
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            (clis[0], ('127.0.0.2', 8082)),
            (clis[1], ('127.0.0.3', 8083)),
            (clis[2], ('127.0.0.4', 8084)),
            TestException(),
            TestException(),
            TestException(),
        ]
        tendrils = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_TCPTendril.side_effect = tendrils[:]
        acceptor = mock.Mock(side_effect=[
            mock.Mock(),
            TestException(),
            mock.Mock(),
        ])
        manager = tcp.TCPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        self.assertFalse(mock_sleep.called)
        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.listen(1024),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.accept(),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        mock_TCPTendril.assert_has_calls([
            mock.call(manager, clis[0], ('127.0.0.2', 8082)),
            mock.call(manager, clis[1], ('127.0.0.3', 8083)),
            mock.call(manager, clis[2], ('127.0.0.4', 8084)),
        ])
        acceptor.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[1]),
            mock.call(tendrils[2]),
        ])
        self.assertFalse(clis[0].close.called)
        clis[1].close.assert_called_once_with()
        self.assertFalse(clis[2].close.called)
        mock_track_tendril.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[2]),
        ])
        tendrils[0]._start.assert_called_once_with()
        self.assertFalse(tendrils[1]._start.called)
        tendrils[2]._start.assert_called_once_with()



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import unittest

import gevent
from gevent import event
from gevent import socket
import mock

from tendril import application
from tendril import connection
from tendril import manager
from tendril import udp


class TestException(Exception):
    pass


class TestUDPTendril(unittest.TestCase):
    @mock.patch.object(connection.Tendril, '_send_streamify',
                       return_value='frame')
    def test_send_frame(self, mock_send_streamify):
        sock = mock.Mock()
        tend = udp.UDPTendril(mock.Mock(sock=sock), 'local_addr',
                              'remote_addr')

        tend.send_frame('a frame')

        mock_send_streamify.assert_called_once_with('a frame')
        sock.sendto.assert_called_once_with('frame', 'remote_addr')

    @mock.patch.object(connection.Tendril, '_send_streamify',
                       return_value='frame')
    def test_send_frame_no_sock(self, mock_send_streamify):
        tend = udp.UDPTendril(mock.Mock(sock=None), 'local_addr',
                              'remote_addr')

        self.assertRaises(ValueError, tend.send_frame, 'a frame')

    @mock.patch.object(connection.Tendril, '_send_streamify',
                       return_value='frame')
    def test_send_frame_bad_sendto(self, mock_send_streamify):
        sock = mock.Mock(**{'sendto.side_effect': TestException()})
        tend = udp.UDPTendril(mock.Mock(sock=sock), 'local_addr',
                              'remote_addr')

        tend.send_frame('a frame')

        mock_send_streamify.assert_called_once_with('a frame')
        sock.sendto.assert_called_once_with('frame', 'remote_addr')

    @mock.patch.object(connection.Tendril, 'close')
    def test_close(self, mock_super_close):
        tend = udp.UDPTendril(mock.Mock(), 'local_addr', 'remote_addr')

        tend.close()

        mock_super_close.assert_called_once_with()


@mock.patch.dict(manager.TendrilManager._managers)
class TestUDPTendrilManager(unittest.TestCase):
    @mock.patch.object(manager.TendrilManager, '__init__')
    def test_init(self, mock_init):
        manager = udp.UDPTendrilManager()

        mock_init.assert_called_once_with(None)
        self.assertEqual(manager._sock, None)
        self.assertIsInstance(manager._sock_event, event.Event)
        self.assertEqual(manager._sock_event.is_set(), False)

    @mock.patch.object(manager.TendrilManager, 'start')
    def test_start(self, mock_start):
        manager = udp.UDPTendrilManager()
        manager._sock = 'sock'
        manager._sock_event = mock.Mock()

        manager.start('acceptor', 'wrapper')

        mock_start.assert_called_once_with('acceptor', 'wrapper')
        self.assertEqual(manager._sock, None)
        manager._sock_event.clear.assert_called_once_with()

    @mock.patch.object(manager.TendrilManager, 'stop')
    def test_stop(self, mock_stop):
        manager = udp.UDPTendrilManager()
        manager._sock = 'sock'
        manager._sock_event = mock.Mock()

        manager.stop('thread')

        mock_stop.assert_called_once_with('thread')
        self.assertEqual(manager._sock, None)
        manager._sock_event.clear.assert_called_once_with()

    @mock.patch.object(manager.TendrilManager, 'shutdown')
    def test_shutdown(self, mock_shutdown):
        manager = udp.UDPTendrilManager()
        manager._sock = 'sock'
        manager._sock_event = mock.Mock()

        manager.shutdown()

        mock_shutdown.assert_called_once_with()
        self.assertEqual(manager._sock, None)
        manager._sock_event.clear.assert_called_once_with()

    @mock.patch.object(manager.TendrilManager, 'connect')
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    @mock.patch.object(udp.UDPTendrilManager, 'local_addr', ('0.0.0.0', 8880))
    def test_connect(self, mock_UDPTendril, mock_track_tendril, mock_connect):
        acceptor = mock.Mock()
        manager = udp.UDPTendrilManager()

        tend = manager.connect(('127.0.0.1', 8080), acceptor)

        mock_connect.assert_called_once_with(('127.0.0.1', 8080), acceptor,
                                             None)
        mock_UDPTendril.assert_called_once_with(manager, ('0.0.0.0', 8880),
                                                ('127.0.0.1', 8080))
        acceptor.assert_called_once_with(mock_UDPTendril.return_value)
        mock_track_tendril.assert_called_once_with(
            mock_UDPTendril.return_value)
        self.assertEqual(id(tend), id(mock_UDPTendril.return_value))

    @mock.patch.object(manager.TendrilManager, 'connect')
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    @mock.patch.object(udp.UDPTendrilManager, 'local_addr', ('0.0.0.0', 8880))
    def test_connect_rejected(self, mock_UDPTendril, mock_track_tendril,
                              mock_connect):
        acceptor = mock.Mock(side_effect=application.RejectConnection())
        manager = udp.UDPTendrilManager()

        tend = manager.connect(('127.0.0.1', 8080), acceptor)

        mock_connect.assert_called_once_with(('127.0.0.1', 8080), acceptor,
                                             None)
        mock_UDPTendril.assert_called_once_with(manager, ('0.0.0.0', 8880),
                                                ('127.0.0.1', 8080))
        acceptor.assert_called_once_with(mock_UDPTendril.return_value)
        self.assertFalse(mock_track_tendril.called)
        self.assertEqual(tend, None)

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'recvfrom.side_effect': TestException(),
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    def test_listener_creator(self, mock_UDPTendril, mock_track_tendril,
                              mock_socket):
        acceptor = mock.Mock()
        manager = udp.UDPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_UDPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'recvfrom.side_effect': TestException(),
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    @mock.patch.object(udp.UDPTendrilManager, 'recv_bufsize', 8192)
    def test_listener_creator_recv_bufsize(self, mock_UDPTendril,
                                           mock_track_tendril, mock_socket):
        acceptor = mock.Mock()
        manager = udp.UDPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.recvfrom(8192),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_UDPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'recvfrom.side_effect': gevent.GreenletExit(),
        'getsockname.return_value': ('127.0.0.1', 8080),
        'close.side_effect': socket.error(),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    def test_listener_killed(self, mock_UDPTendril, mock_track_tendril,
                             mock_socket):
        acceptor = mock.Mock()
        manager = udp.UDPTendrilManager()
        manager.running = True

        with self.assertRaises(gevent.GreenletExit):
            manager.listener(acceptor, None)

        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.recvfrom(4096),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_UDPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'recvfrom.side_effect': TestException(),
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    def test_listener_wrapper(self, mock_UDPTendril, mock_track_tendril,
                              mock_socket):
        wrapped_sock = mock.Mock(**{
            'recvfrom.side_effect': TestException(),
        })
        wrapper = mock.Mock(return_value=wrapped_sock)
        acceptor = mock.Mock()
        manager = udp.UDPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, wrapper)

        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
        ])
        wrapper.assert_called_once_with(mock_socket.return_value)
        wrapped_sock.assert_has_calls([
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_UDPTendril.called)
        self.assertFalse(acceptor.called)
        self.assertFalse(mock_track_tendril.called)

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    def test_listener_noacceptor(self, mock_UDPTendril, mock_track_tendril,
                                 mock_socket):
        msgs = ['msg1', 'msg2', 'msg3']
        mock_socket.return_value.recvfrom.side_effect = [
            (msgs[0], ('127.0.0.2', 8082)),
            (msgs[1], ('127.0.0.3', 8083)),
            (msgs[2], ('127.0.0.4', 8084)),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
        ]
        tend = mock.Mock()
        manager = udp.UDPTendrilManager()
        manager.running = True
        manager.tendrils[(('127.0.0.1', 8080), ('127.0.0.3', 8083))] = tend

        with self.assertRaises(TestException):
            manager.listener(None, None)

        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_UDPTendril.called)
        self.assertFalse(mock_track_tendril.called)
        tend._recv_frameify.assert_called_once_with('msg2')

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    def test_listener_recv_frameify_error(self, mock_UDPTendril,
                                          mock_track_tendril, mock_socket):
        mock_socket.return_value.recvfrom.side_effect = [
            ('frame', ('127.0.0.3', 8083)),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
        ]
        tend = mock.Mock(**{'_recv_frameify.side_effect': TestException()})
        manager = udp.UDPTendrilManager()
        manager.running = True
        manager.tendrils[(('127.0.0.1', 8080), ('127.0.0.3', 8083))] = tend

        with self.assertRaises(TestException):
            manager.listener(None, None)

        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        self.assertFalse(mock_UDPTendril.called)
        self.assertFalse(mock_track_tendril.called)
        tend._recv_frameify.assert_called_once_with('frame')
        tend.close.assert_called_once_with()
        self.assertEqual(tend.closed.call_count, 1)

        args = tend.closed.call_args[0]
        self.assertEqual(len(args), 1)
        self.assertIsInstance(args[0], TestException)

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    def test_listener_acceptor(self, mock_UDPTendril, mock_track_tendril,
                               mock_socket):
        msgs = ['msg1', 'msg2', 'msg3']
        mock_socket.return_value.recvfrom.side_effect = [
            (msgs[0], ('127.0.0.2', 8082)),
            (msgs[1], ('127.0.0.3', 8083)),
            (msgs[2], ('127.0.0.4', 8084)),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
        ]
        tendrils = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_UDPTendril.side_effect = [tendrils[0], tendrils[2]]
        acceptor = mock.Mock()
        manager = udp.UDPTendrilManager()
        manager.running = True
        manager.tendrils[(('127.0.0.1', 8080),
                          ('127.0.0.3', 8083))] = tendrils[1]

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        mock_UDPTendril.assert_has_calls([
            mock.call(manager, ('127.0.0.1', 8080), ('127.0.0.2', 8082)),
            mock.call(manager, ('127.0.0.1', 8080), ('127.0.0.4', 8084)),
        ])
        acceptor.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[2]),
        ])
        mock_track_tendril.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[2]),
        ])
        tendrils[0]._recv_frameify.assert_called_once_with('msg1')
        tendrils[1]._recv_frameify.assert_called_once_with('msg2')
        tendrils[2]._recv_frameify.assert_called_once_with('msg3')

    @mock.patch.object(socket, 'socket', return_value=mock.Mock(**{
        'getsockname.return_value': ('127.0.0.1', 8080),
    }))
    @mock.patch.object(manager.TendrilManager, '_track_tendril')
    @mock.patch.object(udp, 'UDPTendril', return_value=mock.Mock())
    def test_listener_rejector(self, mock_UDPTendril, mock_track_tendril,
                               mock_socket):
        mock_socket.return_value.recvfrom.side_effect = [
            ('msg1', ('127.0.0.2', 8082)),
            ('msg2', ('127.0.0.3', 8083)),
            ('msg3', ('127.0.0.4', 8084)),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
            TestException(),
        ]
        tendrils = [mock.Mock(), mock.Mock(), mock.Mock()]
        mock_UDPTendril.side_effect = tendrils[:]
        acceptor = mock.Mock(side_effect=application.RejectConnection())
        manager = udp.UDPTendrilManager()
        manager.running = True

        with self.assertRaises(TestException):
            manager.listener(acceptor, None)

        mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)
        mock_socket.return_value.assert_has_calls([
            mock.call.bind(('', 0)),
            mock.call.getsockname(),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.recvfrom(4096),
            mock.call.close(),
        ])
        self.assertEqual(manager.local_addr, ('127.0.0.1', 8080))
        mock_UDPTendril.assert_has_calls([
            mock.call(manager, ('127.0.0.1', 8080), ('127.0.0.2', 8082)),
            mock.call(manager, ('127.0.0.1', 8080), ('127.0.0.3', 8083)),
            mock.call(manager, ('127.0.0.1', 8080), ('127.0.0.4', 8084)),
        ])
        acceptor.assert_has_calls([
            mock.call(tendrils[0]),
            mock.call(tendrils[1]),
            mock.call(tendrils[2]),
        ])
        self.assertFalse(mock_track_tendril.called)
        self.assertFalse(tendrils[0]._recv_frameify.called)
        self.assertFalse(tendrils[1]._recv_frameify.called)
        self.assertFalse(tendrils[2]._recv_frameify.called)

    def test_sock_getter_notrunning(self):
        manager = udp.UDPTendrilManager()
        manager._sock = 'socket'
        manager._sock_event = mock.Mock()

        result = manager.sock

        self.assertEqual(result, None)
        self.assertFalse(manager._sock_event.wait.called)

    def test_sock_getter(self):
        manager = udp.UDPTendrilManager()
        manager.running = True
        manager._sock = 'socket'
        manager._sock_event = mock.Mock()

        result = manager.sock

        self.assertEqual(result, 'socket')
        manager._sock_event.wait.assert_called_once_with()



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import gevent
from gevent import event
from gevent import queue
from gevent import socket

from tendril import application
from tendril import connection
from tendril import manager
from tendril import utils


class UDPTendril(connection.Tendril):
    """
    Manages state associated with a single UDP "connection".
    """

    proto = 'udp'

    def send_frame(self, frame):
        """
        Sends a frame to the other end of the connection.
        """

        # Get the socket
        sock = self.manager.sock

        if not sock:
            raise ValueError("UDPTendrilManager not running")

        # Send the packet
        try:
            sock.sendto(self._send_streamify(frame), self.remote_addr)
        except Exception:
            # We're a best-effort service anyway, so ignore exceptions
            pass

    def close(self):
        """
        Close the connection.
        """

        # Untrack the tendril
        super(UDPTendril, self).close()


class UDPTendrilManager(manager.TendrilManager):
    """
    Manages new connections through a particular endpoint.  Handles
    accepting new connections and creating new outgoing connections.
    This class includes the attribute ``recv_bufsize``, which allows
    tuning of the amount of data requested from the system for each
    call to the socket ``recvfrom()`` method.  This class also
    provides the property ``sock``, allowing access to the underlying
    ``socket`` object (or its wrapper).
    """

    proto = 'udp'
    recv_bufsize = 4096

    def __init__(self, endpoint=None):
        """
        Initialize a TendrilManager.

        :param endpoint: Identifies the endpoint of the
                         TendrilManager.  This will be the IP address
                         and port number, as a tuple, on which to
                         accept connections and from which to initiate
                         connections.  If not given, defaults to
                         ``('', 0)``.

        Note: TendrilManager instances can only support one of IPv4 or
        IPv6, and outgoing connection addresses must match how the
        TendrilManager was initialized.  To get a TendrilManager
        supporting outgoing IPv6 connections, use the endpoint
        ``('::', 0)``.
        """

        super(UDPTendrilManager, self).__init__(endpoint)

        # Need a place to store the socket
        self._sock = None
        self._sock_event = event.Event()
        self._sock_event.clear()

    def start(self, acceptor=None, wrapper=None):
        """
        Starts the TendrilManager.

        :param acceptor: If given, specifies a callable that will be
                         called with each newly received Tendril;
                         that callable is responsible for initial
                         acceptance of the connection and for setting
                         up the initial state of the connection.  If
                         not given, no new connections will be
                         accepted by the TendrilManager.
        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.

        For passing extra arguments to the acceptor or the wrapper,
        see the ``TendrilPartial`` class; for chaining together
        multiple wrappers, see the ``WrapperChain`` class.
        """

        super(UDPTendrilManager, self).start(acceptor, wrapper)

        # Reset the socket and socket event
        self._sock = None
        self._sock_event.clear()

    def stop(self, *args):
        """
        Stops the TendrilManager.  Requires cooperation from the
        listener implementation, which must watch the ``running``
        attribute and ensure that it stops accepting connections
        should that attribute become False.  Note that some tendril
        managers will not exit from the listening thread until all
        connections have been closed.
        """

        super(UDPTendrilManager, self).stop(*args)

        # Reset the socket and socket event
        self._sock = None
        self._sock_event.clear()

    def shutdown(self):
        """
        Unconditionally shuts the TendrilManager down, killing all
        threads and closing all tendrils.
        """

        super(UDPTendrilManager, self).shutdown()

        # Reset the socket and socket event
        self._sock = None
        self._sock_event.clear()

    def connect(self, target, acceptor):
        """
        Initiate a connection from the tendril manager's endpoint.
        Once the connection is completed, a UDPTendril object will be
        created and passed to the given acceptor.

        :param target: The target of the connection attempt.
        :param acceptor: A callable which will initialize the state of
                         the new UDPTendril object.
        """

        # Call some common sanity-checks
        super(UDPTendrilManager, self).connect(target, acceptor, None)

        # Construct the Tendril
        tend = UDPTendril(self, self.local_addr, target)

        try:
            # Set up the application
            tend.application = acceptor(tend)
        except application.RejectConnection:
            # The acceptor raised a RejectConnection
            return None

        # OK, let's track the tendril
        self._track_tendril(tend)

        # Might as well return the tendril, too
        return tend

    def listener(self, acceptor, wrapper):
        """
        Listens for new connections to the manager's endpoint.  Once a
        new connection is received, a UDPTendril object is generated
        for it and it is passed to the acceptor, which must initialize
        the state of the connection.  If no acceptor is given, no new
        connections can be initialized.

        :param acceptor: If given, specifies a callable that will be
                         called with each newly received UDPTendril;
                         that callable is responsible for initial
                         acceptance of the connection and for setting
                         up the initial state of the connection.  If
                         not given, no new connections will be
                         accepted by the UDPTendrilManager.
        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.
        """

        # OK, set up the socket
        sock = socket.socket(self.addr_family, socket.SOCK_DGRAM)

        with utils.SocketCloser(sock):
            # Bind to our endpoint
            sock.bind(self.endpoint)

            # Get the assigned port number
            self.local_addr = sock.getsockname()

            # Call any wrappers
            if wrapper:
                sock = wrapper(sock)

        # Senders need the socket, too...
        self._sock = sock
        self._sock_event.set()

        # OK, now go into the listening loop with an error threshold
        # of 10
        closer = utils.SocketCloser(sock, 10,
                                    ignore=[application.RejectConnection])
        while True:
            with closer:
                data, addr = sock.recvfrom(self.recv_bufsize)

                # Look up the tendril or create a new one
                try:
                    tend = self[(self.local_addr, addr)]
                except KeyError:
                    if not acceptor:
                        # Can't accept new connections
                        continue

                    # Construct a Tendril
                    tend = UDPTendril(self, self.local_addr, addr)

                    # Set up the application
                    tend.application = acceptor(tend)

                    # OK, let's track the tendril
                    self._track_tendril(tend)

                # We now have a tendril; process the received data
                try:
                    tend._recv_frameify(data)
                except Exception as exc:
                    # Close the Tendril
                    tend.close()

                    # Notify the application what happened
                    tend.closed(exc)

    @property
    def sock(self):
        """
        Retrieve the socket the UDPTendrilManager is listening on.
        """

        # If we're not running, just return None
        if not self.running:
            return None

        # OK, we're running; wait on the _sock_event
        self._sock_event.wait()

        # Return the socket
        return self._sock



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import abc


__all__ = ["Application", "RejectConnection"]


class RejectConnection(Exception):
    """
    Exception that can be raised by an acceptor to signal that the
    connection is rejected and should be closed.
    """

    pass


class Application(object):
    """
    Base class for tracking application state.  Application classes
    are responsible for implementing the base methods documented here.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, parent):
        """
        Initialize the Application.
        """

        self.parent = parent

    def close(self):
        """
        Close the connection.
        """

        self.parent.close()

    def send_frame(self, frame):
        """
        Send a frame across a connection.
        """

        self.parent.send_frame(frame)

    def closed(self, error):
        """
        Called to notify the application that the connection has been
        closed.  Not called if the application initiates the closure.

        :param error: The exception resulting in the connection
                      closure.  If the closure is due to an EOF
                      condition, will be ``None``.
        """

        pass

    @abc.abstractmethod
    def recv_frame(self, frame):
        """
        Called to pass received frames to the application.

        Note: if this method raises any exceptions, the connection
        will be closed and the ``closed()`` method will be called with
        the exception that was raised.  Also note that this method is
        called in the receiving thread, meaning each frame will be
        processed in turn; if it is desired to process frames in
        parallel, the application should trigger a thread for each
        message, or push frames into a work queue, or otherwise
        arrange for the frame processing to occur in parallel.
        """

        pass  # Pragma: nocover



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import unittest

import mock

from tendril import application
from tendril import connection
from tendril import framers


class TestException(Exception):
    pass


class TendrilForTest(connection.Tendril):
    proto = 'test'

    def send_frame(self, frame):
        pass

    def close(self):
        super(TendrilForTest, self).close()


class ApplicationForTest(application.Application):
    def recv_frame(self, frame):
        pass


class TestTendril(unittest.TestCase):
    def test_init(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        self.assertEqual(tend.manager, 'manager')
        self.assertEqual(tend.local_addr, 'local')
        self.assertEqual(tend.remote_addr, 'remote')
        self.assertEqual(tend._application, None)
        self.assertIsInstance(tend._send_framer, framers.IdentityFramer)
        self.assertIsInstance(tend._recv_framer, framers.IdentityFramer)
        self.assertEqual(id(tend._send_framer), id(tend._recv_framer))
        self.assertIsInstance(tend._send_framer_state, framers.FrameState)
        self.assertIsInstance(tend._recv_framer_state, framers.FrameState)

    @mock.patch.object(TendrilForTest, 'default_framer',
                       new=framers.LineFramer)
    def test_init_alt_framer(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        self.assertEqual(tend.manager, 'manager')
        self.assertEqual(tend.local_addr, 'local')
        self.assertEqual(tend.remote_addr, 'remote')
        self.assertEqual(tend._application, None)
        self.assertIsInstance(tend._send_framer, framers.LineFramer)
        self.assertIsInstance(tend._recv_framer, framers.LineFramer)
        self.assertEqual(id(tend._send_framer), id(tend._recv_framer))
        self.assertIsInstance(tend._send_framer_state, framers.FrameState)
        self.assertIsInstance(tend._recv_framer_state, framers.FrameState)

    def test_wrap(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        self.assertRaises(NotImplementedError, tend.wrap, 'wrapper')

    def test_closed_noapp(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        tend.closed('error')

        # Verifying that it doesn't raise an exception

    def test_closed_withapp(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._application = mock.Mock()

        tend.closed('error')

        tend._application.closed.assert_called_once_with('error')

    def test_closed_witherrorapp(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._application = mock.Mock(**{
            'closed.side_effect': TestException()
        })

        tend.closed('error')

        tend._application.closed.assert_called_once_with('error')

    def test_tendril_key(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        self.assertEqual(tend._tendril_key, ('local', 'remote'))

    def test_send_framer_getter(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        self.assertEqual(id(tend.send_framer), id(tend._send_framer))

    def test_send_framer_setter_badsubclass(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        send_framer = tend._send_framer

        with self.assertRaises(ValueError):
            tend.send_framer = mock.Mock()

        self.assertEqual(id(send_framer), id(tend._send_framer))

    def test_send_framer_setter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        sub_framer = framers.LineFramer()

        tend.send_framer = sub_framer

        self.assertEqual(id(sub_framer), id(tend._send_framer))

    def test_send_framer_deleter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._send_framer = 'framer'

        del tend.send_framer

        self.assertIsInstance(tend._send_framer, framers.IdentityFramer)

    @mock.patch.object(TendrilForTest, 'default_framer',
                       new=framers.LineFramer)
    def test_send_framer_deleter_alt_framer(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._send_framer = 'framer'

        del tend.send_framer

        self.assertIsInstance(tend._send_framer, framers.LineFramer)

    def test_send_framer_state(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        self.assertEqual(id(tend.send_framer_state),
                         id(tend._send_framer_state))

    def test_recv_framer_getter(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        self.assertEqual(id(tend.recv_framer), id(tend._recv_framer))

    def test_recv_framer_setter_badsubclass(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        recv_framer = tend._recv_framer

        with self.assertRaises(ValueError):
            tend.recv_framer = mock.Mock()

        self.assertEqual(id(recv_framer), id(tend._recv_framer))

    def test_recv_framer_setter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        sub_framer = framers.LineFramer()

        tend.recv_framer = sub_framer

        self.assertEqual(id(sub_framer), id(tend._recv_framer))

    def test_recv_framer_deleter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._recv_framer = 'framer'

        del tend.recv_framer

        self.assertIsInstance(tend._recv_framer, framers.IdentityFramer)

    @mock.patch.object(TendrilForTest, 'default_framer',
                       new=framers.LineFramer)
    def test_recv_framer_deleter_alt_framer(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._recv_framer = 'framer'

        del tend.recv_framer

        self.assertIsInstance(tend._recv_framer, framers.LineFramer)

    def test_recv_framer_state(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        self.assertEqual(id(tend.recv_framer_state),
                         id(tend._recv_framer_state))

    def test_framers_getter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._send_framer = 'send'
        tend._recv_framer = 'recv'

        framers = tend.framers

        self.assertIsInstance(framers, connection.TendrilFramers)
        self.assertEqual(framers, ('send', 'recv'))

    def test_framers_setter_tuple(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        send = framers.IdentityFramer()
        recv = framers.IdentityFramer()

        tend.framers = (send, recv)

        self.assertEqual(id(tend._send_framer), id(send))
        self.assertEqual(id(tend._recv_framer), id(recv))

    def test_framers_setter_badtuple(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        framer = framers.IdentityFramer()

        with self.assertRaises(ValueError):
            tend.framers = (framer, framer, framer)
        with self.assertRaises(ValueError):
            tend.framers = (framer,)

        self.assertNotEqual(id(tend._send_framer), id(framer))
        self.assertNotEqual(id(tend._recv_framer), id(framer))

    def test_framers_setter_tuple_badframers(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        framer = framers.IdentityFramer()

        with self.assertRaises(ValueError):
            tend.framers = ('send', framer)
        with self.assertRaises(ValueError):
            tend.framers = (framer, 'recv')

        self.assertNotEqual(id(tend._send_framer), id(framer))
        self.assertNotEqual(id(tend._recv_framer), id(framer))

    def test_framers_setter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        framer = framers.IdentityFramer()

        tend.framers = framer

        self.assertEqual(id(tend._send_framer), id(framer))
        self.assertEqual(id(tend._recv_framer), id(framer))

    def test_framers_setter_badframer(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        framer = mock.Mock()

        with self.assertRaises(ValueError):
            tend.framers = framer

        self.assertNotEqual(id(tend._send_framer), id(framer))
        self.assertNotEqual(id(tend._recv_framer), id(framer))

    def test_framers_deleter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._send_framer = 'framer'
        tend._recv_framer = 'framer'

        del tend.framers

        self.assertIsInstance(tend._send_framer, framers.IdentityFramer)
        self.assertIsInstance(tend._recv_framer, framers.IdentityFramer)
        self.assertEqual(id(tend._send_framer), id(tend._recv_framer))

    @mock.patch.object(TendrilForTest, 'default_framer',
                       new=framers.LineFramer)
    def test_framers_deleter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._send_framer = 'framer'
        tend._recv_framer = 'framer'

        del tend.framers

        self.assertIsInstance(tend._send_framer, framers.LineFramer)
        self.assertIsInstance(tend._recv_framer, framers.LineFramer)
        self.assertEqual(id(tend._send_framer), id(tend._recv_framer))

    def test_framer_states(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._send_framer_state = 'send'
        tend._recv_framer_state = 'recv'

        states = tend.framer_states

        self.assertIsInstance(states, connection.TendrilFramerStates)
        self.assertEqual(states, ('send', 'recv'))

    def test_application_getter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._application = 'application'

        self.assertEqual(tend.application, 'application')

    def test_application_setter_none(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._application = 'application'

        tend.application = None

        self.assertEqual(tend._application, None)

    def test_application_setter_badapp(self):
        tend = TendrilForTest('manager', 'local', 'remote')

        with self.assertRaises(ValueError):
            tend.application = 'application'

        self.assertEqual(tend._application, None)

    def test_application_setter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        app = ApplicationForTest(tend)
        tend.application = app

        self.assertEqual(id(tend._application), id(app))

    def test_application_deleter(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._application = 'application'

        del tend.application

        self.assertEqual(tend._application, None)

    def test_send_streamify(self):
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._send_framer = mock.Mock(
            **{'streamify.return_value': "streamified"})
        tend._send_framer_state = mock.Mock()

        result = tend._send_streamify('frame')

        tend._send_framer_state._reset.assert_called_once_with(
            tend._send_framer)
        tend._send_framer.streamify.assert_called_once_with(
            tend._send_framer_state, 'frame')
        self.assertEqual(result, 'streamified')

    def test_recv_frameify(self):
        generator = mock.Mock(**{'next.side_effect': ['frame1', 'frame2',
                                                      'frame3', 'frame4',
                                                      StopIteration]})
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._recv_framer_state = mock.Mock()
        tend._recv_framer = mock.Mock(**{'frameify.return_value': generator})
        tend._application = mock.Mock()

        tend._recv_frameify("this is a test")

        tend._recv_framer_state._reset.assert_called_once_with(
            tend._recv_framer)
        tend._recv_framer.frameify.assert_called_once_with(
            tend._recv_framer_state, 'this is a test')
        self.assertEqual(generator.next.call_count, 5)
        tend._application.recv_frame.assert_has_calls([
            mock.call('frame1'),
            mock.call('frame2'),
            mock.call('frame3'),
            mock.call('frame4'),
        ])

    def test_recv_frameify_noapplication(self):
        generator = mock.Mock(**{'next.side_effect': ['frame1', 'frame2',
                                                      'frame3', 'frame4',
                                                      StopIteration]})
        tend = TendrilForTest('manager', 'local', 'remote')
        tend._recv_framer_state = mock.Mock()
        tend._recv_framer = mock.Mock(**{'frameify.return_value': generator})

        tend._recv_frameify("this is a test")

        tend._recv_framer_state._reset.assert_called_once_with(
            tend._recv_framer)
        tend._recv_framer.frameify.assert_called_once_with(
            tend._recv_framer_state, 'this is a test')
        self.assertEqual(generator.next.call_count, 5)

    def test_recv_frameify_switch(self):
        class Switcher(object):
            def __init__(inst, tend, framers):
                inst.tendril = tend
                inst.framers = framers
                inst.frames = []

            def recv_frame(inst, frame):
                inst.frames.append(frame)
                if frame in inst.framers:
                    inst.tendril._recv_framer = inst.framers[frame]

        generators = [
            mock.Mock(**{'next.side_effect': ['frame1', 'frame2',
                                              'switch1', 'badframe',
                                              StopIteration],
                         'throw.side_effect': StopIteration}),
            mock.Mock(**{'next.side_effect': ['frame3', 'frame4',
                                              'switch2', 'badframe',
                                              StopIteration],
                         'throw.side_effect': StopIteration}),
            mock.Mock(**{'next.side_effect': ['frame5', 'frame6',
                                              'frame7', 'frame8',
                                              StopIteration],
                         'throw.side_effect': StopIteration}),
        ]

        framers = dict(('switch%d' % i,
                        mock.Mock(**{'frameify.return_value': gen}))
                       for i, gen in enumerate(generators))

        tend = TendrilForTest('manager', 'local', 'remote')
        tend._recv_framer_state = mock.Mock()
        tend._recv_framer = framers['switch0']
        tend._application = Switcher(tend, framers)

        tend._recv_frameify("this is a test")

        tend._recv_framer_state._reset.assert_has_calls([
            mock.call(framers['switch0']),
            mock.call(framers['switch1']),
            mock.call(framers['switch2']),
        ])
        framers['switch0'].frameify.assert_called_once_with(
            tend._recv_framer_state, 'this is a test')
        framers['switch1'].frameify.assert_called_once_with(
            tend._recv_framer_state, '')
        framers['switch2'].frameify.assert_called_once_with(
            tend._recv_framer_state, '')
        self.assertEqual(tend._application.frames, [
            'frame1', 'frame2', 'switch1',
            'frame3', 'frame4', 'switch2',
            'frame5', 'frame6', 'frame7', 'frame8',
        ])

    def test_close(self):
        tend = TendrilForTest(mock.Mock(), 'local', 'remote')

        tend.close()

        tend.manager._untrack_tendril.assert_called_once_with(tend)



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import collections

from gevent import socket

import netaddr


__all__ = ["TendrilPartial", "WrapperChain", "addr_info", "SocketCloser"]


class TendrilPartial(object):
    """
    Similar to ``functools.partial()``; however, the positional
    arguments are passed after the positional arguments to the
    callable invocation.  This is formulated to allow additional
    arguments to be passed to socket wrappers and to connection
    acceptors.
    """

    def __init__(self, func, *args, **kwargs):
        """
        Initialize a TendrilPartial.

        :param func: The function to be called.
        :param args: Extra positional arguments to be passed after the
                     positional arguments in the function call.
        :param kwargs: Extra keyword arguments to be passed in the
                       function call.
        """

        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args):
        """
        Call the actual function.  The passed-in positional arguments
        will be followed by positional and keyword arguments
        identified when the ``TendrilPartial`` object was initialized.
        """

        positional = args + self.args
        return self.func(*positional, **self.kwargs)


class WrapperChain(object):
    """
    Allows multiple wrappers to be chained.  Each wrapper will be
    called in turn, in the order defined, to effect its respective
    changes on the socket.
    """

    def __init__(self, wrapper=None, *args, **kwargs):
        """
        Initialize the WrapperChain.  If a wrapper is given, it will
        be added to the chain as the first wrapper to be invoked.  Any
        extra positional or keyword arguments will be passed to that
        wrapper through construction of a ``TendrilPartial``.
        """

        self._wrappers = []

        # Use the chain method to add the wrapper for us
        if wrapper:
            self.chain(wrapper, *args, **kwargs)

    def __call__(self, sock):
        """
        Call the wrapper chain.  Each wrapper will be called in turn
        on the return result of the previous wrapper.  The result of
        the final wrapper will be returned as the wrapped socket.
        """

        for wrapper in self._wrappers:
            sock = wrapper(sock)

        return sock

    def chain(self, wrapper, *args, **kwargs):
        """
        Add a wrapper to the chain.  Any extra positional or keyword
        arguments will be passed to that wrapper through construction
        of a ``TendrilPartial``.  For convenience, returns the
        WrapperChain object, allowing ``chain()`` to be called on the
        return result to register multiple wrappers.
        """

        if args or kwargs:
            wrapper = TendrilPartial(wrapper, *args, **kwargs)

        self._wrappers.append(wrapper)

        # For convenience...
        return self


def addr_info(addr):
    """
    Interprets an address in standard tuple format to determine if it
    is valid, and, if so, which socket family it is.  Returns the
    socket family.
    """

    # If it's a string, it's in the UNIX family
    if isinstance(addr, basestring):
        return socket.AF_UNIX

    # Verify that addr is a tuple
    if not isinstance(addr, collections.Sequence):
        raise ValueError("address is not a tuple")

    # Make sure it has at least 2 fields
    if len(addr) < 2:
        raise ValueError("cannot understand address")

    # Sanity-check the port number
    if not (0 <= addr[1] < 65536):
        raise ValueError("cannot understand port number")

    # OK, first field should be an IP address; suck it out...
    ipaddr = addr[0]

    # Empty string means IPv4
    if not ipaddr:
        if len(addr) != 2:
            raise ValueError("cannot understand address")

        return socket.AF_INET

    # See if it's valid...
    if netaddr.valid_ipv6(ipaddr):
        if len(addr) > 4:
            raise ValueError("cannot understand address")

        return socket.AF_INET6
    elif netaddr.valid_ipv4(ipaddr):
        if len(addr) != 2:
            raise ValueError("cannot understand address")

        return socket.AF_INET

    raise ValueError("cannot understand address")


class SocketCloser(object):
    """
    Context manager that ensures a socket is closed.
    """

    def __init__(self, sock, err_thresh=0, ignore=None):
        """
        Initialize the SocketCloser.

        :param sock: The socket to be closed in the event of an error.
        :param err_thresh: The maximum number of errors to allow
                           before the socket is closed.  If not given
                           or given as 0, no error counting behavior
                           will be used.
        """

        self.sock = sock
        self.err_thresh = err_thresh
        self.errors = 0 if err_thresh > 0 else None
        self.ignore = frozenset(ignore) if ignore else frozenset()

    def __enter__(self):
        """Enter the context.  Returns the SocketCloser."""

        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """
        Exit the context.  If an exception was raised, ensures that
        the socket is closed.  If the error threshold logic is enabled
        by passing a non-zero value for err_thresh to the initializer,
        socket will only be closed if the exception is not an
        Exception (i.e., a BaseException) or if the error threshold is
        exceeded.
        """

        if exc_type:
            # Skip ignored errors
            if exc_type in self.ignore:
                return True

            # An error occurred; handle threshold counting logic: if
            # we're doing error threshold and the error is not a
            # BaseException (like GreenletExit) and the count of
            # errors is less than the error threshold, we declare the
            # exception handled.
            if (self.errors is not None and
                    issubclass(exc_type, Exception) and
                    self.errors < self.err_thresh):
                # Increment the error count
                self.errors += 1

                return True

            # OK, a legitimate error occurred; ensure the socket gets
            # closed
            try:
                self.sock.close()
            except Exception:
                pass
        elif self.errors:
            # No error occurred, so decrement the error count
            self.errors -= 1



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import abc
import collections

from tendril import application
from tendril import framers


__all__ = ["TendrilFramers", "TendrilFramerStates"]


TendrilFramers = collections.namedtuple('TendrilFramers', ['send', 'recv'])
TendrilFramerStates = collections.namedtuple('TendrilFramerStates',
                                             ['send', 'recv'])


class Tendril(object):
    """
    Manages state associated with a single logical connection, called
    a "tendril".  This is an abstract base class; see the tcp.py and
    udp.py files for implementations.

    Several attributes are available on all subclasses.  They
    include::

    ``manager``
      The responsible TendrilManager.

    ``endpoint``
      The address of the local socket, as a (host, port) tuple.

    ``addr``
      The address of the remote socket, as a (host, port) tuple.

    ``application``
      The application object.  May be changed by the application.
      Should be a subclass of ``tendril.Application``.

    ``recv_framer``
      An instance of a class which chops a received stream into a
      logical sequence of frames, buffering any incomplete frames.

    ``send_framer``
      An instance of a class which assembles sent frames into a
      stream, buffering any incomplete packets.

    ``proto``
      The name of the underlying network protocol.

    ``error``
      An exception object describing the most recent error observed on
      the tendril.  Will be cleared once accessed.  Will be None in
      the event the connection was closed by the peer.
    """

    __metaclass__ = abc.ABCMeta

    default_framer = framers.IdentityFramer

    def __init__(self, manager, local_addr, remote_addr):
        """
        Initialize a Tendril.

        :param manager: The TendrilManager responsible for the
                        Tendril.
        :param local_addr: The address of the local end of the
                           connection represented by the Tendril.
        :param remote_addr: The address of the remote end of the
                            connection represented by the Tendril.
        """

        self.manager = manager
        self.local_addr = local_addr
        self.remote_addr = remote_addr

        self._application = None

        # Set the initial framer
        f = self.default_framer()
        self._send_framer = f
        self._recv_framer = f

        # Set up state for the framer
        self._send_framer_state = framers.FrameState()
        self._recv_framer_state = framers.FrameState()

    def _send_streamify(self, frame):
        """
        Helper method to streamify a frame.
        """

        # Get the state and framer
        state = self._send_framer_state
        framer = self._send_framer

        # Reset the state as needed
        state._reset(framer)

        # Now pass the frame through streamify() and return the result
        return framer.streamify(state, frame)

    def _recv_frameify(self, data):
        """
        Helper method to frameify a stream.
        """

        # Get the state and framer
        state = self._recv_framer_state
        framer = None

        # Grab off as many frames as we can
        frameify = None
        while True:
            # Check if we need to change framers
            if framer != self._recv_framer:
                # Notify the currently-running framer
                if frameify:
                    try:
                        frameify.throw(framers.FrameSwitch)
                    except StopIteration:
                        pass

                # Set up the new framer
                framer = self._recv_framer
                state._reset(framer)
                frameify = framer.frameify(state, data)
                data = ''  # Now part of the state's buffer

            # Get the next frame
            try:
                frame = frameify.next()
            except StopIteration:
                # OK, we've extracted as many frames as we can
                break

            # OK, send the frame to the application
            if self._application:
                self._application.recv_frame(frame)

    def wrap(self, wrapper):
        """
        Allows the underlying socket to be wrapped, as by an SSL
        connection.  Not implemented by all Tendril subclasses.

        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.
        """

        raise NotImplementedError("Cannot wrap this connection")

    def closed(self, error=None):
        """
        Notify the application that the connection has been closed.

        :param error: The exception which has caused the connection to
                      be closed.  If the connection has been closed
                      due to an EOF, pass ``None``.
        """

        if self._application:
            try:
                self._application.closed(error)
            except Exception:
                # Ignore exceptions from the notification
                pass

    @property
    def _tendril_key(self):
        """
        Retrieve a key which can be used to look up this tendril.
        """

        return self.local_addr, self.remote_addr

    @property
    def send_framer(self):
        """
        Retrieve the framer in use for the sending side of the
        connection.
        """

        return self._send_framer

    @send_framer.setter
    def send_framer(self, value):
        """
        Set the framer in use for the sending side of the connection.
        The framer state will be reset next time the framer is used.
        """

        if not isinstance(value, framers.Framer):
            raise ValueError("framer must be an instance of tendril.Framer")

        self._send_framer = value

    @send_framer.deleter
    def send_framer(self):
        """
        Reset the framer in use for the sending side of the connection
        to be a tendril.IdentityFramer.  The framer state will be
        reset next time the framer is used.
        """

        self._send_framer = self.default_framer()

    @property
    def send_framer_state(self):
        """
        Retrieve the framer state in use for the sending side of the
        connection.
        """

        return self._send_framer_state

    @property
    def recv_framer(self):
        """
        Retrieve the framer in use for the receiving side of the
        connection.
        """

        return self._recv_framer

    @recv_framer.setter
    def recv_framer(self, value):
        """
        Set the framer in use for the receiving side of the
        connection.  The framer state will be reset next time the
        framer is used.
        """

        if not isinstance(value, framers.Framer):
            raise ValueError("framer must be an instance of tendril.Framer")

        self._recv_framer = value

    @recv_framer.deleter
    def recv_framer(self):
        """
        Reset the framer in use for the receiving side of the
        connection to be a tendril.IdentityFramer.  The framer state
        will be reset next time the framer is used.
        """

        self._recv_framer = self.default_framer()

    @property
    def recv_framer_state(self):
        """
        Retrieve the framer state in use for the receiving side of the
        connection.
        """

        return self._recv_framer_state

    @property
    def framers(self):
        """
        Retrieve the framers in use for the connection.
        """

        return TendrilFramers(self._send_framer, self._recv_framer)

    @framers.setter
    def framers(self, value):
        """
        Set the framers in use for the connection.  The framer states
        will be reset next time their respective framer is used.
        """

        # Handle sequence values
        if isinstance(value, collections.Sequence):
            if len(value) != 2:
                raise ValueError('need exactly 2 values to unpack')
            elif (not isinstance(value[0], framers.Framer) or
                  not isinstance(value[1], framers.Framer)):
                raise ValueError("framer must be an instance of "
                                 "tendril.Framer")

            self._send_framer, self._recv_framer = value

        # If we have a single value, assume it's a framer
        else:
            if not isinstance(value, framers.Framer):
                raise ValueError("framer must be an instance of "
                                 "tendril.Framer")

            self._send_framer = value
            self._recv_framer = value

    @framers.deleter
    def framers(self):
        """
        Reset the framers in use for the connection to be a
        tendril.IdentityFramer.  The framer states will be reset next
        time their respective framer is used.
        """

        f = self.default_framer()
        self._send_framer = f
        self._recv_framer = f

    @property
    def framer_states(self):
        """
        Retrieve the framer states in use for the connection.
        """

        return TendrilFramerStates(self._send_framer_state,
                                   self._recv_framer_state)

    @property
    def application(self):
        """Retrieve the current application."""

        return self._application

    @application.setter
    def application(self, value):
        """Update the application."""

        # Always allow None
        if value is None:
            self._application = None
            return

        # Check that the state is valid
        if not isinstance(value, application.Application):
            raise ValueError("application must be an instance of "
                             "tendril.Application")

        self._application = value

    @application.deleter
    def application(self):
        """Clear the application state."""

        self._application = None

    @abc.abstractmethod
    def send_frame(self, frame):
        """Send a frame on the connection."""

        pass  # Pragma: nocover

    @abc.abstractmethod
    def close(self):
        """
        Close the connection.  Must not call the ``closed()`` method;
        that will be taken care of at a higher layer.
        """

        # Make sure we're untracked
        self.manager._untrack_tendril(self)

    @abc.abstractproperty
    def proto(self):
        """Retrieve the name of the underlying network protocol."""

        pass  # Pragma: nocover



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import unittest

import gevent
from gevent import socket
import mock

from tendril import utils


class TestTendrilPartial(unittest.TestCase):
    def test_init(self):
        partial = utils.TendrilPartial('func', 1, 2, 3, a=4, b=5, c=6)

        self.assertEqual(partial.func, 'func')
        self.assertEqual(partial.args, (1, 2, 3))
        self.assertEqual(partial.kwargs, dict(a=4, b=5, c=6))

    def test_call(self):
        func = mock.Mock(return_value="result")
        partial = utils.TendrilPartial(func, 1, 2, 3, a=4, b=5, c=6)
        result = partial("arg1", "arg2", "arg3")

        self.assertEqual(result, "result")
        func.assert_called_once_with("arg1", "arg2", "arg3", 1, 2, 3,
                                     a=4, b=5, c=6)


class TestWrapperChain(unittest.TestCase):
    def test_init(self):
        chain = utils.WrapperChain()

        self.assertEqual(chain._wrappers, [])

    def test_init_wrapper(self):
        chain = utils.WrapperChain('wrapper')

        self.assertEqual(chain._wrappers, ['wrapper'])

    @mock.patch.object(utils, 'TendrilPartial',
                       new=lambda func, *args, **kwargs: (func, args, kwargs))
    def test_init_wrapper_args(self):
        chain = utils.WrapperChain('wrapper', 1, 2, 3, a=4, b=5, c=6)

        self.assertEqual(chain._wrappers, [
            ('wrapper', (1, 2, 3), dict(a=4, b=5, c=6)),
        ])

    @mock.patch.object(utils, 'TendrilPartial',
                       new=lambda func, *args, **kwargs: (func, args, kwargs))
    def test_chain(self):
        # chain() has already been tested during the tests of
        # __init__() above, so we need merely do one last double-check
        # and test the convenience return
        chain = utils.WrapperChain()

        result = chain.chain('wrapper', 1, 2, 3, a=4, b=5, c=6)

        self.assertEqual(id(chain), id(result))
        self.assertEqual(chain._wrappers, [
            ('wrapper', (1, 2, 3), dict(a=4, b=5, c=6)),
        ])

    def test_call(self):
        chain = utils.WrapperChain(mock.Mock(return_value="sock2"))\
            .chain(mock.Mock(return_value="sock3"))\
            .chain(mock.Mock(return_value="sock4"))

        result = chain("sock1")

        self.assertEqual(result, "sock4")
        chain._wrappers[0].assert_called_once_with("sock1")
        chain._wrappers[1].assert_called_once_with("sock2")
        chain._wrappers[2].assert_called_once_with("sock3")


class TestAddrInfo(unittest.TestCase):
    def test_accept_unix(self):
        result = utils.addr_info('unixaddr')

        self.assertEqual(result, socket.AF_UNIX)

    def test_reject_nonseq(self):
        self.assertRaises(ValueError, utils.addr_info, mock.Mock())

    def test_reject_short(self):
        self.assertRaises(ValueError, utils.addr_info, ('127.0.0.1',))

    def test_reject_badport(self):
        self.assertRaises(ValueError, utils.addr_info, ('127.0.0.1', -1))
        self.assertRaises(ValueError, utils.addr_info, ('127.0.0.1', 65536))

    def test_accept_empty(self):
        result = utils.addr_info(('', 8080))

        self.assertEqual(result, socket.AF_INET)

    def test_reject_empty_too_many_fields(self):
        self.assertRaises(ValueError, utils.addr_info, ('', 8080, 0))

    def test_accept_ipv6(self):
        result = utils.addr_info(('::1', 8080))

        self.assertEqual(result, socket.AF_INET6)

    def test_accept_ipv6_flowinfo(self):
        result = utils.addr_info(('::1', 8080, 0))

        self.assertEqual(result, socket.AF_INET6)

    def test_accept_ipv6_scopeid(self):
        result = utils.addr_info(('::1', 8080, 0, 0))

        self.assertEqual(result, socket.AF_INET6)

    def test_reject_ipv6_too_many_fields(self):
        self.assertRaises(ValueError, utils.addr_info, ('::1', 8080, 0, 0, 0))

    def test_accept_ipv4(self):
        result = utils.addr_info(('127.0.0.1', 8080))

        self.assertEqual(result, socket.AF_INET)

    def test_reject_ipv4_too_many_fields(self):
        self.assertRaises(ValueError, utils.addr_info, ('127.0.0.1', 8080, 0))

    def test_reject_unknown(self):
        self.assertRaises(ValueError, utils.addr_info, ('unknown', 8080))


class TestException(Exception):
    pass


class TestSocketCloser(unittest.TestCase):
    def test_init(self):
        closer = utils.SocketCloser('sock')

        self.assertEqual(closer.sock, 'sock')
        self.assertEqual(closer.err_thresh, 0)
        self.assertEqual(closer.errors, None)
        self.assertEqual(closer.ignore, frozenset())

    def test_init_err_thresh(self):
        closer = utils.SocketCloser('sock', 10)

        self.assertEqual(closer.sock, 'sock')
        self.assertEqual(closer.err_thresh, 10)
        self.assertEqual(closer.errors, 0)
        self.assertEqual(closer.ignore, frozenset())

    def test_init_ignore(self):
        closer = utils.SocketCloser('sock', ignore=[TestException])

        self.assertEqual(closer.sock, 'sock')
        self.assertEqual(closer.err_thresh, 0)
        self.assertEqual(closer.errors, None)
        self.assertEqual(closer.ignore, frozenset([TestException]))

    def test_enter(self):
        closer = utils.SocketCloser('sock')
        result = closer.__enter__()

        self.assertEqual(closer, result)

    def test_exit_noerror(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock)

        result = closer.__exit__(None, None, None)

        self.assertEqual(result, None)
        self.assertFalse(sock.close.called)
        self.assertEqual(closer.errors, None)

    def test_exit_error(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock)

        result = closer.__exit__(TestException, TestException('foo'), [])

        self.assertEqual(result, None)
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, None)

    def test_exit_ignore(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock, ignore=[TestException])

        result = closer.__exit__(TestException, TestException('foo'), [])

        self.assertEqual(result, True)
        self.assertFalse(sock.close.called)
        self.assertEqual(closer.errors, None)

    def test_exit_kill(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock)

        result = closer.__exit__(gevent.GreenletExit, gevent.GreenletExit(),
                                 [])

        self.assertEqual(result, None)
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, None)

    def test_exit_closeerror(self):
        sock = mock.Mock(**{'close.side_effect': socket.error()})
        closer = utils.SocketCloser(sock)

        result = closer.__exit__(TestException, TestException('foo'), [])

        self.assertEqual(result, None)
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, None)

    def test_exit_noerror_thresh(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock, 10)
        closer.errors = 5

        result = closer.__exit__(None, None, None)

        self.assertEqual(result, None)
        self.assertFalse(sock.close.called)
        self.assertEqual(closer.errors, 4)

    def test_exit_error_thresh(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock, 1)

        result = closer.__exit__(TestException, TestException('foo'), [])

        self.assertEqual(result, True)
        self.assertFalse(sock.close.called)
        self.assertEqual(closer.errors, 1)

        result = closer.__exit__(TestException, TestException('foo'), [])

        self.assertEqual(result, None)
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, 1)

    def test_exit_kill_thresh(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock, 10)

        result = closer.__exit__(gevent.GreenletExit, gevent.GreenletExit(),
                                 [])

        self.assertEqual(result, None)
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, 0)

    def test_exit_closeerror_thresh(self):
        sock = mock.Mock(**{'close.side_effect': socket.error()})
        closer = utils.SocketCloser(sock, 1)

        result = closer.__exit__(TestException, TestException('foo'), [])

        self.assertEqual(result, True)
        self.assertFalse(sock.close.called)
        self.assertEqual(closer.errors, 1)

        result = closer.__exit__(TestException, TestException('foo'), [])

        self.assertEqual(result, None)
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, 1)

    def test_together_noerror(self):
        sock = mock.Mock()

        with utils.SocketCloser(sock):
            sock.make_call()

        sock.make_call.assert_called_once_with()
        self.assertFalse(sock.close.called)

    def test_together_error(self):
        sock = mock.Mock()

        with self.assertRaises(TestException):
            with utils.SocketCloser(sock):
                sock.make_call()
                raise TestException('spam')

        sock.make_call.assert_called_once_with()
        sock.close.assert_called_once_with()

    def test_together_kill(self):
        sock = mock.Mock()

        with self.assertRaises(gevent.GreenletExit):
            with utils.SocketCloser(sock):
                sock.make_call()
                raise gevent.GreenletExit()

        sock.make_call.assert_called_once_with()
        sock.close.assert_called_once_with()

    def test_together_closeerror(self):
        sock = mock.Mock(**{'close.side_effect': socket.error()})

        with self.assertRaises(TestException):
            with utils.SocketCloser(sock):
                sock.make_call()
                raise TestException('spam')

        sock.make_call.assert_called_once_with()
        sock.close.assert_called_once_with()

    def test_together_noerror_thresh(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock, 10)
        closer.errors = 5

        with closer:
            sock.make_call()

        sock.make_call.assert_called_once_with()
        self.assertFalse(sock.close.called)
        self.assertEqual(closer.errors, 4)

    def test_together_error_thresh(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock, 1)

        with closer:
            sock.make_call()
            raise TestException('spam')

        sock.make_call.assert_called_once_with()
        self.assertFalse(sock.close.called)
        self.assertEqual(closer.errors, 1)

        sock.reset_mock()

        with self.assertRaises(TestException):
            with closer:
                sock.make_call()
                raise TestException('spam')

        sock.make_call.assert_called_once_with()
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, 1)

    def test_together_kill_thresh(self):
        sock = mock.Mock()
        closer = utils.SocketCloser(sock, 10)

        with self.assertRaises(gevent.GreenletExit):
            with closer:
                sock.make_call()
                raise gevent.GreenletExit()

        sock.make_call.assert_called_once_with()
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, 0)

    def test_together_closeerror_thresh(self):
        sock = mock.Mock(**{'close.side_effect': socket.error()})
        closer = utils.SocketCloser(sock, 1)

        with closer:
            sock.make_call()
            raise TestException('spam')

        sock.make_call.assert_called_once_with()
        self.assertFalse(sock.close.called)
        self.assertEqual(closer.errors, 1)

        sock.reset_mock()

        with self.assertRaises(TestException):
            with closer:
                sock.make_call()
                raise TestException('spam')

        sock.make_call.assert_called_once_with()
        sock.close.assert_called_once_with()
        self.assertEqual(closer.errors, 1)



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import unittest

import mock

from tendril import application


class ApplicationForTest(application.Application):
    def recv_frame(self, frame):
        pass


class TestApplication(unittest.TestCase):
    def test_init(self):
        app = ApplicationForTest('parent')

        self.assertEqual(app.parent, 'parent')

    def test_close(self):
        app = ApplicationForTest(mock.Mock())
        app.close()

        app.parent.close.assert_called_once_with()

    def test_send_frame(self):
        app = ApplicationForTest(mock.Mock())
        app.send_frame('frame')

        app.parent.send_frame.assert_called_once_with('frame')



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

"""
==============================================
Tendril Frame-based Network Connection Tracker
==============================================

Tendril is a network communication library based on two main features:
it is based on sending and receiving frames, and it tracks the state
of an abstract connection as defined by the application.  Tendril is
designed to be easy to use: creating an application requires
subclassing the ``Application`` class and providing an implementation
for the recv_frame() method; then get a ``TendrilManager`` class
instance and start it, and Tendril manages the rest.

Tendril Concepts
================

Frames
------

Tendril is based on the concept of passing around frames or packets of
data.  The fact is, most network protocols are based on sending and
receiving frames; for instance, in the SMTP protocol used for sending
email, the sender will start off sending the frame "MAIL FROM
email@example.com" followed by a line termination sequence (a carriage
return followed by a newline).  The SMTP server will then respond with
another frame acknowledging the "MAIL FROM" frame, and that frame will
also end with a line termination sequence.  Thus, even though SMTP is
defined on top of the TCP protocol, which provides an undivided stream
of data between the client and server, a framing boundary is imposed
upon it--in this case, the carriage return followed by a newline that
terminates each frame.

Tendril includes the concept of *framers*.  A framer is nothing more
than a subclass of ``Framer`` which has one method which extracts a
single frame from the stream of undifferentiated data, and another
method which converts a frame into an appropriate representation.  In
the case of the SMTP protocol exchange above, the ``frameify()``
method finds each line terminated by the carriage return-newline pair,
strips off those characters, and returns just the frame.  In the same
way, the corresponding ``streamify()`` method takes the frame and
appends a carriage return-newline pair.

For text-based protocols such as SMTP, this may seem like overkill.
However, for binary-based protocols, a lot of code is dedicated to
determining the boundaries between frames, and in some cases even
decoding the frame.  Tendril's concept of a framer for a connection
enables the framing logic to be isolated from the rest of the
application, and even reused: Tendril comes with several pre-built
framers, including framers designed to work with a text-based protocol
such as SMTP.

Another important advantage of the framer concept is the ability to
switch between framers as needed.  Taking again the example of the
SMTP protocol--the actual email data is transferred to the server by
the client first sending a "DATA" frame; the server responds
indicating that it is ready to begin receiving the message data, and
then the client simply sends the message data, ending it with a line
containing only a single period (".").  In this case, an SMTP server
application based on Tendril may wish to receive the message data as a
single frame; it can do this by creating a framer which buffers stream
data until it sees that ending sentinel (the period on a line by
itself), then returns the whole message as a single frame.  Once the
server receives the "DATA" frame from the client, all it has to do is
temporarily switch out the framer in use for the receiving side of the
connection, then switch it back to the standard line-based framer once
it has received the message frame.

Tendril allows for different framers to be used on the receiving side
and sending side of the connection.  This could be used in a case like
the SMTP server example cited above, where the server still wishes to
send line-oriented frames to the client, even while buffering a
message data frame.  In addition, although the provided framers deal
with byte data, Tendril itself treats the frames as opaque;
applications can use this to build a framer that additionally parses a
given frame into a class object that the rest of the application then
processes as necessary.

Connection Tracking
-------------------

Tendril is also based on the concept of tracking connection state.
For connection-oriented protocols such as TCP, obviously, this is not
a big problem; however, Tendril is also designed to support
connectionless protocols such as UDP, where some applications need to
manage state information relevant to a given exchange.  As an
admittedly contrived example, consider DNS, which is based on UDP.  A
client of the DNS system will send a request to a DNS server over UDP;
when a response is received from that DNS server, the connection state
information tracked by Tendril can help connect that response with the
appropriate request, ensuring that the response goes to the right
place.

This connection state tracking is primarily intended to assist
applications which desire to be available over both
connection-oriented protocols such as TCP and over connectionless
protocols such as UDP.  Although Tendril does not address reliability
or frame ordering, its connection state tracking eases the
implementation of an application which utilizes both types of
protocols.

Extensibility
-------------

Careful readers may have noticed the use of the terms, "such as TCP"
and "such as UDP."  Although Tendril only has built-in support for TCP
and UDP connections, it is possible to extend Tendril to support other
protocols.  All that is required is to create subclasses of
``Tendril`` (representing an individual connection) and of
``TendrilManager`` (which accepts and creates connections and manages
any necessary socket data flows), and to register the
``TendrilManager`` subclasses as ``pkg_resources`` entry points under
the ``tendril.manager`` namespace.  See the ``setup.py`` for Tendril
for an example of how this may be done.

In addition to allowing Tendril to support protocols other than TCP
and UDP, it is also possible to implement new framers by subclassing
the ``Framer`` class.  (Note: as Tendril deals with ``Framer``
objects, it is not necessary to register these framers using
``pkg_resources`` entry points.)  Objects of these classes may then
simply be assigned to the appropriate ``framers`` attribute on the
``Tendril`` instance representing the connection.

Advanced Interfaces
-------------------

Tendril also provides an advanced interface that allows a given raw
socket to be "wrapped."  Using this feature, an ordinary TCP socket
could be converted into an SSL socket.  Other uses for this interface
are possible, such as setting socket options for the socket.  Tendril
also provides an interface to allow multiple of these wrapper
functions to be called in a given order.

Standard Usage
==============

The first step in using Tendril is to define an application by
subclassing the ``Application`` class.  (Subclassing is not strictly
necessary--Tendril uses Python's standard ``abc`` package for defining
abstract base classes--but using subclassing will pull in a few
helpful and/or required methods.)  The subclass need merely implement
the recv_frame() method, which will be called when a frame is
received.  The ``Application`` subclass constructor itself can be the
*acceptor* to be used by Tendril (more on acceptors in a moment).

Once the ``Application`` subclass has been created, the developer then
needs to get a ``TendrilManager`` instance, using the
``get_manager()`` factory function.  The exact call to
``get_manager()`` depends on the needs; for making outgoing
connections, simply calling ``get_manager("tcp")`` is sufficient.  If
listening on a port or making an outgoing connection from a specific
address and/or port is desired, the second argument to
``get_manager()`` may be a tuple of the desired local IP address and
the port number (i.e., ``("127.0.0.1", 80)``).

All managers must be started, and ``get_manager()`` does not start the
manager by itself.  Check the manager's ``running`` attribute to see
if the manager is already running, and if it is not, call its
``start()`` method.  To accept connections, pass ``start()`` the
*acceptor* (usually the ``Application`` subclass).  The ``start()``
method also accepts a *wrapper*, which will be called with the
listening socket when it is created.

If, instead of accepting connections (as a server would do), the
desire is to make outgoing connections, simply call ``start()`` with
no arguments, then call the ``connect()`` method of the manager.  This
method takes the *target* of the connection (i.e., the IP address and
port number, as a tuple) and the *acceptor*.  (It also has an optional
*wrapper*, which will be called with the outgoing socket just prior to
initiating the connection.)

Acceptors
---------

An *acceptor* is simply a callable taking a single argument--the
``Tendril`` instance representing the connection--and returning an
instance of a subclass of ``Application``, which will be assigned to
the ``application`` attribute of the ``Tendril`` instance.  The
acceptor initializes the application; it also has the opportunity to
manipulate that ``Tendril``, such as setting framers, calling the
``Tendril`` instance's ``wrap()`` method, or simply closing the
connection.

Although the ``TendrilManager`` does not provide the opportunity to
pass arguments to the acceptor, it is certainly possible to do so.
The standard Python ``functools.partial()`` is one obvious interface,
but Tendril additionally provides its own ``TendrilPartial`` utility;
the advantage of ``TendrilPartial`` is that the positional argument
passed to the acceptor--the ``Tendril`` instance--will be the first
positional argument, rather than the last one, as would be the case
with ``functools.partial()``.

Wrappers
--------

A *wrapper* is simply a callable again taking a single argument--in
this case, the socket object--and returning a wrapped version of that
argument; that wrapped version of the socket will then be used in
subsequent network calls.  A wrapper which manipulates socket options
can simply return the socket object which was passed in, while one
which performs SSL encapsulation can return the SSL wrapper.  Again,
although there is no opportunity to pass arguments to the wrapper in a
manager ``start()`` or ``connect()`` call (or a ``Tendril`` object's
``wrap()`` call), ``functools.partial()`` or Tendril's
``TendrilPartial`` utility can be used.  In particular, in conjunction
with ``TendrilPartial``, the ``ssl.wrap_socket()`` call can be used as
a socket wrapper directly, enabling an SSL connection to be set up
easily.

Of course, it may be necessary to perform multiple "wrapping"
activities on a connection, such as setting socket options followed by
wrapping the socket in an SSL connection.  For this case, Tendril
provides the ``WrapperChain``; it can be initialized in the same way
that ``TendrilPartial`` is, but additional wrappers can be added by
calling the ``chain()`` method; when called, the ``WrapperChain``
object will call each wrapper in the order defined, returning the
final wrapped socket in the end.
"""

from application import *
from connection import *
from framers import *
from manager import *
from utils import *


__all__ = (application.__all__ + connection.__all__ + framers.__all__ +
           manager.__all__ + utils.__all__)
